import  numpy as np
import scipy as sp
import scipy.linalg
import copy
cimport numpy as np
cimport cython
from libc.math cimport sqrt, pow, log, sin, cos, atan2, sqrt
from cython.parallel import prange
cdef double PI = 3.1415926535
from scipy.sparse import spdiags
from scipy.sparse.linalg.eigen.arpack import eigs, ArpackNoConvergence
from scipy.misc import derivative



def parse_model_spec(model_spec, param_keys):

    # First, extract the classes
    class_list = model_spec['classes']
    nClass = len(class_list) # total number of classes

    # Make dictionaries for class and parameter index look-up
    class_index_dict = {class_name:i for (i, class_name) in enumerate(class_list)}
    params_index_dict = {param: i for (i, param) in enumerate(param_keys)}

    try:
        # dictionaries of all constant, linear and all infection terms
        constant_dict = {}
        linear_dict = {}
        infection_dict = {}
        finres_dict = {}

        for class_name in class_list:
            reaction_dict = model_spec[class_name]
            if 'constant' in reaction_dict.keys():
                constant_dict[class_name] = reaction_dict['constant']
            if 'linear' in reaction_dict.keys():
                linear_dict[class_name] = reaction_dict['linear']
            if 'infection' in reaction_dict.keys():
                infection_dict[class_name] = reaction_dict['infection']
            if 'finite-resource' in reaction_dict.keys():
                finres_dict[class_name] = reaction_dict['finite-resource']

        # parse the constant term
        constant_term_set = set() # used to check for duplicates
        constant_term_list = []
        for (k, val) in constant_dict.items():
            for (rate,) in val:
                if (k, rate) in constant_term_set:
                    raise Exception('Duplicate constant term: {}, {}'.format(k, rate))
                else:
                    sign = 1
                    constant_term_set.add((k, rate))
                    class_index = class_index_dict[k]
                    if rate.startswith('-'):
                        rate = rate[1:]
                        sign = -1
                    rate_index = params_index_dict[rate]
                    constant_term_list.append([rate_index, class_index, sign])

        if len(constant_term_list) > 0: # add one class for Ni
            class_index_dict['Ni'] = nClass
            nClass += 1 # add one class for Ni

        # parse the linear terms into a list of [rate_index, reagent_index] and a dictionary for the product
        linear_terms_set = set() # used to check for duplicates
        linear_terms_list = [] # collect all linear terms
        linear_terms_destination_dict = {} # a dictionary for the product
        for (k, val) in linear_dict.items():
            for (reagent, rate) in val:
                if (reagent, rate) in linear_terms_set:
                    raise Exception('Duplicates linear terms: {}, {}'.format(reagent, rate))
                else:
                    linear_terms_set.add((reagent, rate))
                    reagent_index = class_index_dict[reagent]
                    if rate.startswith('-'):
                        rate = rate[1:]
                        rate_index = params_index_dict[rate]
                        linear_terms_list.append([rate_index, reagent_index, -1])
                    else:
                        rate_index = params_index_dict[rate]
                        linear_terms_destination_dict[(rate_index, reagent_index)] = class_index_dict[k]

        # parse the infection terms into a list of [rate_index, reagent_index] and a dictionary for the product
        infection_terms_set = set() # used to check to duplicates
        infection_terms_list = [] # collect all infection terms
        infection_terms_destination_dict = {} # a dictionary for the product
        for (k, val) in infection_dict.items():
            for (reagent, rate) in val:
                if (reagent, rate) in infection_terms_set:
                    raise Exception('Duplicates infection terms: {}, {}'.format(reagent, rate))
                else:
                    infection_terms_set.add((reagent, rate))
                    reagent_index = class_index_dict[reagent]
                    if rate.startswith('-'):
                        rate = rate[1:]
                        if k != 'S':
                            raise Exception('A susceptible group that is not S: {}'.format(k))
                        else:
                            rate_index = params_index_dict[rate]
                            infection_terms_list.append([rate_index, reagent_index, -1])
                    else:
                        rate_index = params_index_dict[rate]
                        infection_terms_destination_dict[(rate_index, reagent_index)] = class_index_dict[k]


        # parse the finite resource-terms into
        #  1) a list resource_list = [rate_index, (class_index0, priority_index0), (class_index1, priority_index1), ...]
        #  2) a list finres_terms_list = [resource_index, priority_index, positivity_index, class_index, origin_class_index, destination_class_index]

        finres_terms_set = set() # used to check to duplicates
        finres_terms_dict = {} # collect all finite-resource terms
        finres_terms_list = [] # collect all finite-resource terms
        resource_list = []
        resource_dict = {}
        resource_count = 0
        for (k, val) in finres_dict.items():
            for (reagent, rate, priority, prob) in val:
                if (reagent, rate, priority, prob) in finres_terms_set:
                    raise Exception('Duplicates finite-resource terms: {}, {}, {}, {}'.format(reagent, rate, priority, prob))
                else:
                    finres_terms_set.add((reagent, rate, priority, prob))
                    class_index = class_index_dict[reagent]
                    neg_rate = rate.startswith('-')
                    if neg_rate:
                        rate = rate[1:]
                    rate_index = params_index_dict[rate]
                    priority_index = params_index_dict[priority]
                    prob_index = params_index_dict[prob]
                    if rate not in resource_dict.keys():
                        resource_dict[rate] = resource_count
                        resource_list.append([rate_index])
                        resource_count += 1
                    resource_index = resource_dict[rate]
                    if (class_index, priority_index) not in resource_list[resource_index]:
                        resource_list[resource_index].append((class_index, priority_index))
                    dict_key = (resource_index, priority_index, prob_index, class_index)
                    if dict_key not in finres_terms_dict.keys():
                        finres_terms_dict[dict_key] = [-1,-1]
                    if neg_rate:
                        finres_terms_dict[dict_key][0] = class_index_dict[k]
                    else:
                        finres_terms_dict[dict_key][1] = class_index_dict[k]
        for (k,val) in finres_terms_dict.items():
            finres_terms_list.append(list(k) + val)




        # parse parameters for testing (for SppQ only, otherwise ignore empty parameters lists)
        test_pos_list = []
        test_freq_list = []
        if "test_pos" in model_spec and "test_freq" in model_spec:
            test_pos_strings = model_spec['test_pos']
            test_freq_strings = model_spec['test_freq']

            if len(test_pos_strings) != len(class_list)+1 or len(test_freq_strings) != len(class_list)+1:
                raise Exception('Test parameters must be specified for every class (including R). {} parameters expected'.format(len(class_list)+1))

            for rate in test_pos_strings:
                rate_index = params_index_dict[rate]
                test_pos_list.append(rate_index)
            for rate in test_freq_strings:
                rate_index = params_index_dict[rate]
                test_freq_list.append(rate_index)

            for class_name in class_list: # add quarantine classes
                class_index_dict[class_name+'Q'] = nClass
                nClass += 1
            class_index_dict['NiQ'] = nClass
            nClass += 1




    except KeyError:
        raise Exception('No reactions for some classes. Please check model_spec again')

    # set the product index
    set_destination(linear_terms_list, linear_terms_destination_dict)
    set_destination(infection_terms_list, infection_terms_destination_dict)






    res = (nClass, class_index_dict, np.array(constant_term_list, dtype=np.intc, ndmin=2),
                                     np.array(linear_terms_list, dtype=np.intc, ndmin=2),
                                     np.array(infection_terms_list, dtype=np.intc, ndmin=2),
                                     np.array(finres_terms_list, dtype=np.intc, ndmin=2),
                                     np.array(resource_list, dtype=object),
                                     np.array(test_pos_list, dtype=np.intc, ndmin=1),
                                     np.array(test_freq_list, dtype=np.intc, ndmin=1))
    return res

def set_destination(term_list, destination_dict):
    '''
    A function used by parse_model_spec that sets the product_index
    '''
    for term in term_list:
        rate_index = term[0]
        reagent_index = term[1]
        if (rate_index, reagent_index) in destination_dict.keys():
            product_index = destination_dict[(rate_index, reagent_index)]
            term[2] = product_index

def age_dep_rates(rate, int M, str name, bint check_length=True):
    if np.size(rate)==1:
        return rate*np.ones(M)
    elif np.size(rate)==M or check_length==False:
        return rate
    else:
        raise Exception('{} can be a number or an array of size M'.format(name))


DTYPE = np.float
ctypedef np.float_t DTYPE_t

cpdef forward_euler_integration(f, double [:] x, double t1, double t2, Py_ssize_t steps, Py_ssize_t inter_steps=1):
    cdef:
        double dt=(t2-t1)/(inter_steps*(steps-1)),t=t1
        double [:] fx
        Py_ssize_t i, j, size=x.shape[0]
        double [:, :] sol=np.empty((inter_steps*steps, size), dtype=DTYPE)
    for j in range(size):
        sol[0, j] = x[j]
    for i in range(1, inter_steps*steps):
        fx = f(t, sol[i-1])
        for j in range(size):
            sol[i, j] = sol[i-1, j] + fx[j]*dt
        t += dt
    return sol

cpdef RK2_integration(f, double [:] x, double t1, double t2, Py_ssize_t steps):
    cdef:
        double dt=(t2-t1)/(steps-1),t=t1
        double [:] fx
        Py_ssize_t i, j, size=x.shape[0]
        double [:] k1=np.empty((size), dtype=DTYPE), temp=np.empty((size), dtype=DTYPE)
        double [:, :] sol=np.empty((steps, size), dtype=DTYPE)

    for j in range(size):
        sol[0, j] = x[j]
    for i in range(1, steps):
        fx = f(t, sol[i-1])
        for j in range(size):
            k1[j] = dt*fx[j]
            temp[j] = sol[i-1, j] + k1[j]
        t += dt
        fx = f(t, temp)
        for j in range(size):
            sol[i, j] = sol[i-1, j] + 0.5*(k1[j] + dt*fx[j])
    return sol

cpdef nearest_positive_definite(double [:, :] A):
    """Find the nearest positive-definite matrix to input

    [1] https://gist.github.com/fasiha/fdb5cec2054e6f1c6ae35476045a0bbd

    [2] https://www.mathworks.com/matlabcentral/fileexchange/42885-nearestspd

    [3] N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6
    """
    cdef:
        np.ndarray[DTYPE_t, ndim=2] B, V, H, A2, A3, I
        np.ndarray[DTYPE_t, ndim=1] s
        double spacing, mineig
        int k

    B = np.add(A, np.transpose(A))/2
    _, s, V = np.linalg.svd(B)
    H = np.dot(V.T, np.dot(np.diag(s), V))
    A2 = np.add(B, H) / 2
    A3 = (A2 + A2.T) / 2

    if is_positive_definite(A3):
        return A3
    else:
        spacing = np.spacing(np.linalg.norm(B))
        I = np.eye(B.shape[0])
        k = 1
        while not is_positive_definite(A3):
            mineig = np.min(np.real(np.linalg.eigvals(A3)))
            A3 += I * (np.abs(mineig) * k**2 + spacing)
            k += 1
        return A3

cpdef is_positive_definite(double [:, :] B):
    """Returns true when input is positive-definite, via Cholesky"""
    try:
        _ = np.linalg.cholesky(B)
        sign, _ = np.linalg.slogdet(B)
        if sign > 0:
            return True
        else:
            return False
    except np.linalg.LinAlgError:
        return False


cpdef solve_symmetric_close_to_singular(double [:, :] A, double [:] b, double eps=1e-13):
    """ Solve the linear system Ax=b and return x and the log-determinant of A for a
    symmetric matrix A that should be positive definite but might be close to singular
    or slightly non-positive.

    If A is symmteric positive, this function simply solves the linear system and computes
    the determinant using a Cholesky decomposition. Otherwise, it computes the pseudo-inverse
    and pseudo-determinant of A accounting only for positive and large-enough eigenvalues.

    Parameters
    ----------
    A: numpy.array(dims=2)
        The symmetric matrix.
    b: numpy.array(dims=1)
        The right-hand side of the linear equation.
    eps: float, optional
        Any eigenvalue smaller than eps*max_eigenval is ignored for the computation of the
        pseudo-inverse.

    Returns
    -------
    x: numpy.array(dims=1)
        The solution of Ax=b.
    log_det:
        The (pseudo-)log-determinant of A.
    """
    cdef:
        double log_det
        np.ndarray[DTYPE_t, ndim=1] x
    try:
        chol_fact = sp.linalg.cho_factor(A)
        log_det = 2*np.sum(np.log(np.diag(chol_fact[0])))
        x = sp.linalg.cho_solve(chol_fact, b)
        return x, log_det

    except sp.linalg.LinAlgError:
        # Not positive definite:
        eigvals, eigvecs = np.linalg.eigh(A)

        # Set all small or negative eigenvalues to 0 and compute pseudo-inverse and
        # pseudo-determinant.
        eigvals[eigvals < eps] = 0
        log_det = np.sum(np.log(eigvals[eigvals > 0]))

        eigvals[eigvals != 0] = 1/eigvals[eigvals != 0]

        x = eigvecs @ (np.diag(eigvals) @ (eigvecs.T @ b))
        return x, log_det

def largest_real_eig(np.ndarray A):
    w, v = np.linalg.eig(A)
    max_index = np.argmax(np.real(w))
    eigval = np.real(w[max_index])
    eigvec = np.real(v[:, max_index])
    eigval_sign = (eigval > 0)
    return eigval_sign, eigvec



def hessian_finite_difference(pos, function, eps=1e-3, method="central"):
    """Forward finite-difference computation of the Hessian of a function.

    Parameters
    ----------
    pos:numpy.array(dims=1)
        Position at which the hessian is to be computed.
    function: function(numpy.array)
        Function of interest.
    pos: float or numpy.array(dims=1), optional
        Step size used for FD computation (can be parameter dependant).
    method: str
        Different options for the FD computation: "forward" or "central".

    Returns
    -------
    hess: numpy.array(dims=2)
        Hessian of function at pos.
    """
    k = len(pos)
    if not hasattr(eps, "__len__"):
        eps = eps*np.ones(k)

    hessian = np.empty((k, k))

    if method == "forward":
        val_central = function(pos)
        val1 = np.zeros(k)
        for i in range(k):
            pos[i] += eps[i]
            val1[i] = function(pos)
            pos[i] -= eps[i]

        for i in range(k):
            pos[i] += eps[i]
            for j in range(k):
                pos[j] += eps[j]
                val2 = function(pos)
                pos[j] -= eps[j]

                hessian[i, j] = (val2 - val1[i] - val1[j] + val_central)/(eps[i]*eps[j])
            pos[i] -= eps[i]

        return 1/2 * (hessian + hessian.T)

    if method == "central":
        orig_pos = pos.copy()
        for i in range(k):
            for j in range(i+1):
                pos = orig_pos.copy()
                pos[i] += eps[i]
                pos[j] += eps[j]
                val1 = function(pos)
                pos[j] -= 2*eps[j]
                val2 = function(pos)
                pos = orig_pos.copy()
                pos[i] -= eps[i]
                pos[j] += eps[j]
                val3 = function(pos)
                pos[j] -= 2*eps[j]
                val4 = function(pos)
                hessian[i, j] = (val1 + val4 - val2 - val3) / (4*eps[i]*eps[j])
                hessian[j, i] = (val1 + val4 - val2 - val3) / (4*eps[i]*eps[j])

        return hessian

    raise Exception("Finite-difference method must be 'forward' or 'central'.")

def partial_derivative(func, var, point, dx, *func_args):
    args = point[:]
    def wraps(x, *wraps_args):
        args[var] = x
        return func(args, *wraps_args)
    return derivative(wraps, point[var], dx=dx, args=func_args)

cpdef make_fltr(fltr_list, n_list):
    fltr = [f for (i, f) in enumerate(fltr_list) for n in range(n_list[i])]
    return np.array(fltr)

def process_fltr(np.ndarray fltr, Py_ssize_t Nf):
    if fltr.ndim == 2:
        return np.array([fltr]*Nf)
    elif fltr.shape[0] == Nf:
        return fltr
    else:
        raise Exception("fltr must be a 2D array or an array of 2D arrays")

def process_obs(np.ndarray obs, Py_ssize_t Nf):
    if obs.shape[0] != Nf:
        raise Exception("Wrong length of obs")
    if obs.ndim == 2:
        return np.ravel(obs)
    elif obs.ndim == 1:
        return np.concatenate(obs)
    else:
        raise Exception("Obs must be a 2D array or an array of 1D arrays")

def process_latent_data(np.ndarray fltr, np.ndarray obs):
    cdef Py_ssize_t Nf=obs.shape[0]
    fltr = process_fltr(fltr, Nf)
    obs0 = obs[0]
    obs = process_obs(obs[1:], Nf-1)
    return fltr, obs, obs0

def _parse_prior_name(sub_dict, dim):
    if 'prior_fun' in sub_dict:
        prior = sub_dict['prior_fun']
        if np.size(prior) == 1:
            return [prior]*dim
        elif np.size(prior) == dim:
            return prior
        else:
            raise Exception('prior_fun must be of length 1 or {}'.format(dim))
    else:
        return ['lognorm']*dim

def parse_param_prior_dict(prior_dict, M, check_length=True):
    flat_guess = []
    flat_stds = []
    flat_bounds = []
    flat_guess_range = []
    key_list = []
    scaled_guesses = []
    is_scale_parameter = []
    names = []

    count = 0
    for key in prior_dict:
        key_list.append(key)
        sub_dict = prior_dict[key]
        try:
            mean = sub_dict['mean']
        except KeyError:
            raise Exception('Sub-dict under {} must have "mean" as a key'.format(key))
        if np.size(mean) == 1:
            flat_guess.append(mean)
            try:
                flat_stds.append(sub_dict['std'])
                flat_bounds.append(sub_dict['bounds'])
            except KeyError:
                raise Exception('Sub-dict under {} must have "std" and "bounds"'
                                ' as keys'.format(key))
            flat_guess_range.append(count)
            is_scale_parameter.append(False)
            names += _parse_prior_name(sub_dict, 1)
            count += 1
        else:
            infer_scale = False
            if 'infer_scale' in sub_dict.keys():
                infer_scale = sub_dict['infer_scale']
            if infer_scale:
                flat_guess.append(1.0)
                try:
                    flat_stds.append(sub_dict['scale_factor_std'])
                    flat_bounds.append(sub_dict['scale_factor_bounds'])
                except KeyError:
                    raise Exception('Sub-dict under {} must have "scale_factor_std"'
                                    'and "scale_factor_bounds" as keys because'
                                    'infer_scale" is True'.format(key))
                scaled_guesses.append(mean)
                flat_guess_range.append(count)
                is_scale_parameter.append(True)
                names += _parse_prior_name(sub_dict, 1)
                count += 1
            else:
                if check_length:
                    assert len(mean) == M, 'length of mean must be either 1 or M'
                else:
                    M = len(mean)
                flat_guess += list(mean)
                try:
                    assert len(sub_dict['std']) == M
                    assert len(sub_dict['bounds']) == M
                    flat_stds += list(sub_dict['std'])
                    flat_bounds += list(sub_dict['bounds'])
                except KeyError:
                    raise Exception('Sub-dict under {} must have "std" and "bounds"'
                                    ' as keys'.format(key))
                is_scale_parameter += [False]*M
                flat_guess_range.append(list(range(count, count+M)))
                names += _parse_prior_name(sub_dict, M)
                count += M
    return names, key_list, np.array(flat_guess), np.array(flat_stds), \
          np.array(flat_bounds), flat_guess_range, is_scale_parameter, scaled_guesses

def unflatten_parameters(params, flat_guess_range, is_scale_parameter, scaled_guesses):
    # Restore parameters from flattened parameters
    orig_params = []
    k=0
    for j in range(len(flat_guess_range)):
        if is_scale_parameter[j]:
            orig_params.append(np.array([params[flat_guess_range[j]]*val for val in scaled_guesses[k]]))
            k += 1
        else:
            orig_params.append(params[flat_guess_range[j]])
    return orig_params

def parse_init_prior_dict(prior_dict, dim, obs_dim):
    guess = []
    stds = []
    bounds = []
    names = []
    flags = [False, False]
    fltrs = [None, None]
    count = 0
    if 'lin_mode_coeff' in prior_dict.keys():
        sub_dict = prior_dict['lin_mode_coeff']
        try:
            guess.append(sub_dict['mean'])
            stds.append(sub_dict['std'])
            bounds.append(sub_dict['bounds'])
            fltrs[0] = _process_init_fltr(sub_dict['fltr'], dim)
        except KeyError:
            raise Exception('Sub dict of "lin_mode_coeff" must have'
                            ' "mean", "std", "bounds" and "fltr" as keys')
        flags[0] = True
        names += _parse_prior_name(sub_dict, 1)
        count += fltrs[0].shape[0]

    if 'independent' in prior_dict.keys():
        sub_dict = prior_dict['independent']
        try:
            fltrs[1] = _process_init_fltr(sub_dict['fltr'], dim)
            guess.extend(sub_dict['mean'])
            stds.extend(sub_dict['std'])
            bounds.extend(sub_dict['bounds'])
        except KeyError:
            raise Exception('Sub dict of "independent" must have'
                            ' "mean", "std", "bounds" and "fltr" as keys')
        assert fltrs[1].shape[0] == len(sub_dict['mean'])
        assert len(sub_dict['std']) == len(sub_dict['mean'])
        assert len(sub_dict['bounds']) == len(sub_dict['mean'])
        flags[1] = True
        names += _parse_prior_name(sub_dict, len(sub_dict['mean']))
        count += fltrs[1].shape[0]

    # make sure that there are some priors
    if np.sum(flags) == 0:
        raise Exception('Prior for inits must have at least one of "independent"'
                        ' and "coeff" as keys')
    # check that the total number of guesses is correct
    assert count == dim - obs_dim, 'Total No. of guessed values must be dim - obs_dim'

    return names, np.array(guess), np.array(stds), np.array(bounds), \
           flags, fltrs

def _process_init_fltr(fltr, dim):
    fltr = np.array(fltr).astype('bool')
    if fltr.ndim == 1:
        assert len(fltr) == dim, 'init_fltr has wrong shape'
        l = np.sum(fltr)
        fltr_mat = np.zeros((l, dim), dtype=DTYPE)
        row = 0
        for i in range(dim):
            if fltr[i]:
                fltr_mat[row, i] = 1
                row += 1
        return fltr_mat
    elif fltr.ndim == 2:
        assert fltr.shape[1] == dim, 'init_fltr has wrong shape'
        return fltr.astype('float')
    else:
        raise Exception('init_fltr must either be 1D or 2D')

cpdef double distance_on_Earth(double [:] coord1, double [:] coord2):
    cdef:
        double lat1=coord1[0], lon1=coord1[1], lat2=coord2[0], lon2=coord2[1]
        double Earth_radius_km=6371.0, degree_to_radius=PI/180.0
        double d_lat, d_lon, a, c
    d_lat = (lat2 - lat1) * degree_to_radius
    d_lon = (lon2 - lon2) * degree_to_radius

    lat1 *= degree_to_radius
    lat2 *= degree_to_radius

    a = sin(d_lat/2) * sin(d_lat/2) + sin(d_lon/2) * sin(d_lon/2) * cos(lat1) * cos(lat2)
    c = 2 * atan2(sqrt(a), sqrt(1-a))
    return Earth_radius_km * c


def plotSIR(data, showPlot=True):
    t = data['t']
    X = data['X']
    M = data['M']
    Ni = data['Ni']
    N = Ni.sum()

    S = X[:, 0: M]
    Is = X[:, 2*M:3*M]
    R = Ni - X[:, 0:M] - X[:, M:2*M] - X[:, 2*M:3*M]


    sumS = S.sum(axis=1)
    sumI = Is.sum(axis=1)
    sumR = R.sum(axis=1)

    try:
        import matplotlib.pyplot as plt
        plt.fill_between(t, 0, sumS/N, color="#348ABD", alpha=0.3)
        plt.plot(t, sumS/N, '-', color="#348ABD", label='$S$', lw=4)

        plt.fill_between(t, 0, sumI/N, color='#A60628', alpha=0.3)
        plt.plot(t, sumI/N, '-', color='#A60628', label='$I$', lw=4)

        plt.fill_between(t, 0, sumR/N, color="dimgrey", alpha=0.3)
        plt.plot(t, sumR/N, '-', color="dimgrey", label='$R$', lw=4)

        plt.legend(fontsize=26); plt.grid()
        plt.autoscale(enable=True, axis='x', tight=True)

        plt.ylabel('Fraction of compartment value')
        plt.xlabel('Days')
        if True != showPlot:
            pass
        else:
            plt.show()

    except ImportError:
        print('Please install matplotlib to use this method')




def getPopulation(country='India', M=16):
    """
    Takes coutry name and number, M, of age-groups as argument
    Returns population structured in M age-groups

    Parameters
    ----------
    country: string
        Default is 'India'
    M: int
        Deafault is 16 age-groups
    """
    u0 = 'https://raw.githubusercontent.com/rajeshrinet/pyross/master/examples/data/'

    u1 = u0 + 'age_structures/India-2019.csv'
    u2 = u0 + 'age_structures/UK.csv'
    u3 = u0 + 'age_structures/Germany-2019.csv'
    u4 = u0 + 'age_structures/Italy-2019.csv'
    u5 = u0 + 'age_structures/Denmark-2019.csv'
    u6 = u0 + 'age_structures/UK.csv'
    u7 = u0 + 'age_structures/US.csv'
    u8 = u0 + 'age_structures/China-2019.csv'
    u9 = u0 + 'age_structures/France.txt'

    if country=='India':
        data = np.asanyarray(np.genfromtxt(u1, delimiter=','))
        N_m  = np.array((data[:, 1]))[1:M+1]
        N_f  = np.array((data[:, 2 ]))[1:M+1]
        Ni   = N_m + N_f 
        Ni   = Ni[0:M];  Ni=Ni.astype('double')

    elif country=='UK':
        data = np.asanyarray(np.genfromtxt(u2, delimiter=','))
        N_m  = np.array((data[:, 1]))[1:M+1]
        N_f  = np.array((data[:, 2 ]))[1:M+1]
        Ni   = N_m + N_f 
        Ni   = Ni[0:M];  Ni=Ni.astype('double')

    elif country=='Germany':
        data = np.asanyarray(np.genfromtxt(u3, delimiter=','))
        N_m  = np.array((data[:, 1]))[1:M+1]
        N_f  = np.array((data[:, 2 ]))[1:M+1]
        Ni   = N_m + N_f 
        Ni   = Ni[0:M];  Ni=Ni.astype('double')

    elif country=='Italy':
        data = np.asanyarray(np.genfromtxt(u4, delimiter=','))
        N_m  = np.array((data[:, 1]))[1:M+1]
        N_f  = np.array((data[:, 2 ]))[1:M+1]
        Ni   = N_m + N_f 
        Ni   = Ni[0:M];  Ni=Ni.astype('double')

    elif country=='Denmark':
        data = np.asanyarray(np.genfromtxt(u5, delimiter=','))
        N_m  = np.array((data[:, 1]))[1:M+1]
        N_f  = np.array((data[:, 2 ]))[1:M+1]
        Ni   = N_m + N_f 
        Ni   = Ni[0:M];  Ni=Ni.astype('double')

    elif country=='UK':
        data = np.asanyarray(np.genfromtxt(u6, delimiter=','))
        N_m  = np.array((data[:, 1]))[1:M+1]
        N_f  = np.array((data[:, 2 ]))[1:M+1]
        Ni   = N_m + N_f 
        Ni   = Ni[0:M];  Ni=Ni.astype('double')

    elif country=='USA':
        data = np.asanyarray(np.genfromtxt(u7, delimiter=','))
        N_m  = np.array((data[:, 1]))[1:M+1]
        N_f  = np.array((data[:, 2 ]))[1:M+1]
        Ni   = N_m + N_f 
        Ni   = Ni[0:M];  Ni=Ni.astype('double')

    elif country=='China':
        data = np.asanyarray(np.genfromtxt(u8, delimiter=','))
        N_m  = np.array((data[:, 1]))[1:M+1]
        N_f  = np.array((data[:, 2 ]))[1:M+1]
        Ni   = N_m + N_f 
        Ni   = Ni[0:M];  Ni=Ni.astype('double')

    elif country=='France':
        Ni = np.genfromtxt(u9)[:, 3]

    else:
        print('Direct extraction of Ni is not implemnted , please do it locally')

    return Ni


def get_summed_CM(CH0, CW0, CS0, CO0, M, M0, Ni, Ni0):
    CH = np.zeros((M, M))
    CW = np.zeros((M, M))
    CS = np.zeros((M, M))
    CO = np.zeros((M, M))

    for i in range(16):
        CH0[i,:] = Ni0[i]*CH0[i,:]
        CS0[i,:] = Ni0[i]*CS0[i,:]
        CW0[i,:] = Ni0[i]*CW0[i,:]
        CO0[i,:] = Ni0[i]*CO0[i,:]

    for i in range(M):
        for j in range(M):
            i1, j1 = i*M, j*M
            CH[i,j] = np.sum( CH0[i1:i1+M,j1:j1+M] )/Ni[i]
            CW[i,j] = np.sum( CW0[i1:i1+M,j1:j1+M] )/Ni[i]
            CS[i,j] = np.sum( CS0[i1:i1+M,j1:j1+M] )/Ni[i]
            CO[i,j] = np.sum( CO0[i1:i1+M,j1:j1+M] )/Ni[i]
    return CH, CW, CS, CO


class GPR:
    def __init__(self, nS, nT, iP, nP, xS, xT, yT):
        self.nS   =  nS           # # of test data points
        self.nT   =  nT           # # of training data points
        self.iP   =  iP           # # inverse of sigma
        self.nP   =  nP           # # number of priors
        self.xS   =  xS           # test input
        self.xT   =  xT           # training input
        self.yT   =  yT           # training output

        self.yS   =  0            # test output
        self.yP   =  0            # prior output
        self.K    =  0            # kernel
        self.Ks   =  0            # kernel
        self.Kss  =  0            # kernel
        self.mu   =  0            # mean
        self.sd   =  0            # stanndard deviation


    def calcDistM(self, r, s):
        '''Calculate distance matrix between 2 1D arrays'''
        return r[..., np.newaxis] - s[np.newaxis, ...]


    def calcKernels(self):
        '''Calculate the kernel'''
        cc = self.iP*0.5
        self.K   = np.exp(-cc*self.calcDistM(self.xT, self.xT)**2)
        self.Ks  = np.exp(-cc*self.calcDistM(self.xT, self.xS)**2)
        self.Kss = np.exp(-cc*self.calcDistM(self.xS, self.xS)**2)
        return


    def calcPrior(self):
        '''Calculate the prior'''
        L  = np.linalg.cholesky(self.Kss + 1e-6*np.eye(self.nS))
        G  = np.random.normal(size=(self.nS, self.nP))
        yP = np.dot(L, G)
        return


    def calcMuSigma(self):
        '''Calculate the mean'''
        self.mu =  np.dot(self.Ks.T, np.linalg.solve(self.K, self.yT))

        vv = self.Kss - np.dot(self.Ks.T, np.linalg.solve(self.K, self.Ks))
        self.sd = np.sqrt(np.abs(np.diag(vv)))

        # Posterior
        L       = np.linalg.cholesky(vv + 1e-6*np.eye(self.nS))
        self.yS = self.mu.reshape(-1,1) + np.dot(L, np.random.normal(size=(self.nS, self.nP)))
        return


    def plotResults(self):
        try:
            import matplotlib.pyplot as plt
            plt.plot(self.xT, self.yT, 'o', ms=10, mfc='#348ABD', mec='none', label='training set' )
            plt.plot(self.xS, self.yS, '#dddddd', lw=1.5, label='posterior')
            plt.plot(self.xS, self.mu, '#A60628', lw=2, label='mean')

            # fill 95% confidence interval (2*sd about the mean)
            plt.fill_between(self.xS.flat, self.mu-2*self.sd, self.mu+2*self.sd, color="#348ABD", alpha=0.4, label='2 sigma')
            plt.axis('tight'); plt.legend(fontsize=15); plt.rcParams.update({'font.size':18})
        except ImportError:
            print('Please install matplotlib to use this method')


    def runGPR(self):
        self.calcKernels()
        self.calcPrior()
        self.calcMuSigma()
        self.plotResults()


def resample(weighted_samples, N):
    """
    Given a set of weighted samples, produce a set of unweighted samples approximating the same
    distribution. We implement residual resampling here, see https://doi.org/10.1109/ISPA.2005.195385
    for some context. This function can (and for large N will) duplicate likely samples.

    Parameters
    ----------
    weighted_samples: list of dict
        The set of weighed samples (e.g. given by nested sampling).
    N: int
        The number of samples that should be drawn from the given samples.

    Returns
    -------
    sample_list: list of dict
        A set of uniformly weighted samples.
    """

    weights = np.array([w['weight'] for w in weighted_samples])
    weights /= np.sum(weights)

    # Deterministic part
    selected_samples = np.array([int(w*N) for w in weights])
    # Random part
    selected_samples += np.random.multinomial(N - sum(selected_samples), weights, size=1)[0,:]

    sample_list = []
    for i in range(len(selected_samples)):
        for j in range(selected_samples[i]):
            sample_list.append(weighted_samples[i])

    return sample_list


def posterior_mean(weighted_samples):
    """
    Compute the posterior mean of a set of (potentially) weighted samples of the posterior
    (e.g. computed by nested sampling or by MCMC).

    Parameters
    ----------
    weighted_samples: list of dict
        The set of (weighted) samples.

    Returns
    -------
    sample: dict
        An approximation of the posterior mean.
    """
    weights = np.array([w['weight'] for w in weighted_samples])
    weights /= np.sum(weights)

    sample = weighted_samples[0].copy()
    sample['weight'] = 1.0

    # Set the parameter to averages.
    potential_list_keys = {'params_dict', 'control_params_dict',  # Current keys
                           'map_params_dict', 'map_dict' }        # Keys from deprecated interface (to be deleted)
    potential_keys = {'flat_params', 'x0',   # Current keys
                      'flat_map', 'map_x0'}  # Keys from deprecated interface (to be deleted)

    # Select the keys that are actually in the result dictionaries.
    list_keys = list(potential_list_keys & set(sample.keys()))
    keys = list(potential_keys & set(sample.keys()))

    # Set the average parameters
    for list_key in list_keys:
        for key in sample[list_key].keys():
            vals = [w[list_key][key] for w in weighted_samples]
            avg = sum([weights[i] * vals[i] for i in range(len(vals))])
            sample[list_key][key] = avg

    for key in keys:
        vals = [w[key] for w in weighted_samples]
        avg = sum([weights[i] * vals[i] for i in range(len(vals))])
        sample[key] = avg

    # Reset the value of log-posterior (we cannot compute this here)
    sample['log_posterior']  = None
    sample['log_prior']      = None
    sample['log_likelihood'] = None
    return sample


def build_SppQ_model_spec(input_model_spec, testRate_name = "tau"):
    """
    Adds a quarantined version of every state and automatically specifies finite-resource 
    transtions for testing. If not yet specified, an explicit R-class is added. 

    Parameters
    inout_model_spec: dict
        Model specification of the model without quarantine (Spp syntax)
  
    Returns
    -------
    model_spec: dict
        Full model specification with quarantine.
    """
    
    model_spec = copy.deepcopy(input_model_spec)
    
    # collect all terms
    terms = list()
    for name in model_spec['classes']:
        for key, listval in model_spec[name].items():
            for val in listval:
                terms.append((key,val))
            
    
    # Add R-class, if not yet specified
    if "R" not in model_spec['classes']:
        model_spec['classes'].append("R")
        model_spec['R'] = dict()
        # check if transition is balanced, otherwise add to R
        for term in terms:
            term2 = copy.deepcopy(term)
            if term[1][1].startswith('-'):
                term2[1][1] = term[1][1][1:]
                if term2 not in terms:
                    if term2[0] not in model_spec['R'].keys():
                        model_spec['R'][term2[0]] = list()
                    model_spec['R'][term2[0]].append(term2[1])
    n_class = len(model_spec['classes'])
                    
    # add Q-classes
    for class_name in model_spec['classes'].copy():
        classQ = class_name + 'Q'
        model_spec['classes'].append(classQ)
        model_spec[classQ] = copy.deepcopy(model_spec[class_name])
        for terms in model_spec[classQ].values():
            for term in terms:
                term[0] += 'Q'
        if 'infection' in model_spec[classQ]:
            del model_spec[classQ]['infection'] # no infection in quarantine
    
    # add testing transitions
    test_pos = model_spec.pop('test_pos')
    assert len(test_pos) == n_class, "number of elements of test_pos must match number of classes (incl. R)"
    test_freq = model_spec.pop('test_freq')
    assert len(test_freq) == n_class, "number of elements of test_freq must match number of classes (incl. R)"
    
    for i in range(n_class):
        class_name = model_spec['classes'][i]
        class_nameQ = model_spec['classes'][i+n_class]
        model_spec[class_name]['finite-resource'] = [ [class_name, "-"+testRate_name, test_freq[i], test_pos[i]] ]
        model_spec[class_nameQ]['finite-resource'] = [ [class_name, testRate_name, test_freq[i], test_pos[i]] ]
        
                
    return model_spec



import pyross.contactMatrix
import pyross.control
import pyross.deterministic
import pyross.forecast
import pyross.hybrid
import pyross.inference
import pyross.stochastic
import pyross.utils
import pyross.utils_python
import pyross.evidence
import pyross.tsi.deterministic
import pyross.tsi.utils

## current version of PyRoss
__version__="1.3.4"




import  numpy as np 
cimport numpy as np
cimport cython
from libc.stdlib cimport malloc, free
import warnings
from scipy.special import legendre
import matplotlib.pyplot as plt
from scipy.integrate import odeint
import nlopt

DTYPE   = np.float


cdef class CommonMethods:
    """
    Parent class used for all classes listed below. 
    It includes:
    a) Integrators used by various deterministic models listed below.
    b) Method to get time series of S, etc by passing a dict of data.
    c) Method to set the contactMatrix array, CM
    """

    def simulator(self, x0, contactMatrix, Tf, Nf, integrator='odeint', 
        Ti=0, maxNumSteps=100000, **kwargs):
        """
        Simulates a compartment model given initial conditions,
        choice of integrator and other parameters. 
        Returns the time series data and parameters in a dict. 
        
        ...

        Parameters
        ----------
        x0: np.array
            Initial state vector (number of compartment values).
            An array of size M*(model_dimension-1),
            where x0[i+j*M] should be the initial
            value of model class i of age group j.
            The removed R class must be left out.
            If Ni is dynamical, then the last M points store Ni.
        contactMatrix: python function(t)
             The social contact matrix C_{ij} denotes the
             average number of contacts made per day by an
             individual in class i with an individual in class j
        Tf: float
            Final time of integrator
        Nf: Int
            Number of time points to evaluate.
        Ti: float, optional
            Start time of integrator. The default is 0.
        integrator: TYPE, optional
            Integrator to use either from scipy.integrate or odespy.
            The default is 'odeint'.
        maxNumSteps: int, optional
            maximum number of steps the integrator can take.
            The default is 100000.
        **kwargs: kwargs for integrator

        Returns
        -------
        data: dict
             X: output path from integrator,  t : time points evaluated at,
            'param': input param to integrator.

        """

        def dxdtEval(xt, t):
            self.CM = contactMatrix(t)
            self.rhs(xt, t)
            return self.dxdt

        if integrator=='odeint':
            from scipy.integrate import odeint
            time_points=np.linspace(Ti, Tf, Nf)   
            X = odeint(dxdtEval, x0, time_points, mxstep=maxNumSteps, **kwargs)

        elif integrator=='solve_ivp':
            from scipy.integrate import solve_ivp
            time_points=np.linspace(Ti, Tf, Nf)                                                          
            X = solve_ivp(lambda t, xt: dxdtEval(xt,t), [Ti,Tf], x0, 
                         t_eval=time_points, **kwargs).y.T

        elif integrator=='odespy' or integrator=='odespy-vode':
            import odespy
            time_points=np.linspace(Ti, Tf, Nf)                                                          
            solver = odespy.Vode(dxdtEval, method = 'bdf', 
                    atol=1E-7, rtol=1E-6, order=5, nsteps=maxNumSteps)
            solver.set_initial_condition(x0)
            X, time_points = solver.solve(time_points, **kwargs)

        elif integrator=='odespy-rkf45':
            import odespy
            time_points=np.linspace(Ti, Tf, Nf)                                                          
            solver = odespy.RKF45(dxdtEval)
            solver.set_initial_condition(x0)
            X, time_points = solver.solve(time_points, **kwargs)

        elif integrator=='odespy-rk4':
            import odespy
            time_points=np.linspace(Ti, Tf, Nf)
            solver = odespy.RK4(dxdtEval)
            solver.set_initial_condition(x0)
            X, time_points = solver.solve(time_points, **kwargs)

        else:
            raise Exception("Error: Integration method not found! \n \
                            Please set integrator='odeint' to use \n \
                            the scipy.integrate's odeint (Default)\n \
                            Use integrator='odespy-vode' to use vode \
                            from odespy (github.com/rajeshrinet/odespy).\n \
                            Use integrator='odespy-rkf45' to use RKF45  \
                            from odespy (github.com/rajeshrinet/odespy).\n \
                            Use integrator='odespy-rk4' to use RK4 from \
                            odespy (github.com/rajeshrinet/odespy).     \n \
                            Alternatively, write your own integrator to \
                            evolve the system in time and store the data.\n")

        data     = {'X':X, 't':time_points, 'Ni':self.Ni, 'M':self.M}
        data_out = data.copy()
        data_out.update(self.paramList)
        return data_out

    
    cpdef set_contactMatrix(self, double t, contactMatrix):
        self.CM=contactMatrix(t)


    def S(self,  data):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             S: Susceptible population time series
        """
        
        X = data['X']
        S = X[:, 0:self.M]
        return S


    def E(self,  data, Ei=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             E: Exposed population time series
        """
        
        if None != Ei:
            X = data['X']  
            E = X[:, Ei[0]*self.M:Ei[1]*self.M]
        else:
            X = data['X']  
            Ei=self.readData['Ei']
            E = X[:, Ei[0]*self.M:Ei[1]*self.M]
        return E


    def A(self,  data, Ai=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             A: Activated population time series
        """
        
        if None != Ai:
            X = data['X']  
            A = X[:, Ai[0]*self.M:Ai[1]*self.M]
        else:
            X = data['X']  
            Ai=self.readData['Ai']
            A = X[:, Ai[0]*self.M:Ai[1]*self.M]
        return A


    def I(self,  data, Ii=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Ia : Asymptomatics population time series
        """

        if None != Ii:
            X  = data['X']
            Ii=self.readData['Ii']
            I = X[:, Ii[0]*self.M:Ii[1]*self.M]
        else:
            X  = data['X']  
            Ii=self.readData['Ii']
            I = X[:, Ii[0]*self.M:Ii[1]*self.M]
        return I


    def Ia(self,  data, Iai=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Ia : Asymptomatics population time series
        """
        
        if None != Iai:
            X  = data['X']
            Ia = X[:, Iai[0]*self.M:Iai[1]*self.M]
        else:
            X  = data['X'] 
            Iai=self.readData['Iai']
            Ia = X[:, Iai[0]*self.M:Iai[1]*self.M]
        return Ia


    def Is(self,  data, Isi=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Is : symptomatics population time series
        """
        if None != Isi:
            X  = data['X']
            Is = X[:, Isi[0]*self.M:Isi[1]*self.M]
        else:
            X  = data['X']  
            Isi=self.readData['Isi']
            Is = X[:, Isi[0]*self.M:Isi[1]*self.M]
        return Is


    def Isp(self,  data, Ispi=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Isp : (intermediate stage between symptomatics 
                   and recovered) population time series
        """
        if None != Ispi:
            X  = data['X']
            Isp = X[:, Ispi[0]*self.M:Ispi[1]*self.M]
        else:
            X  = data['X']  
            Ispi=self.readData['Ispi']
            Isp = X[:, Ispi[0]*self.M:Ispi[1]*self.M]
        return Isp


    def Ih(self,  data, Ihi=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Ic : hospitalized population time series
        """
        if None != Ihi:
            X  = data['X']
            Ih = X[:, Ihi[0]*self.M:Ihi[1]*self.M]
        else:
            X  = data['X']
            Ihi=self.readData['Ihi']
            Ih = X[:, Ihi[0]*self.M:Ihi[1]*self.M]
        return Ih


    def Ihp(self,  data, Ihpi=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Ihp : (intermediate stage between symptomatics 
                   and recovered) population time series
        """
        if None != Ihpi:
            X  = data['X']
            Ihp = X[:, Ihpi[0]*self.M:Ihpi[1]*self.M]
        else:
            X  = data['X']
            Ihpi=self.readData['Ihpi']
            Ihp = X[:, Ihpi[0]*self.M:Ihpi[1]*self.M]
        return Ihp


    def Ic(self,  data, Ici=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Ic : ICU hospitalized population time series
        """
        if None != Ici:
            X  = data['X'] 
            Ic = X[:, Ici[0]*self.M:Ici[1]*self.M ]
        else:
            X  = data['X'] 
            Ici=self.readData['Ici']
            Ic = X[:, Ici[0]*self.M:Ici[1]*self.M ]
        return Ic


    def Icp(self,  data, Icpi=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Icp : (intermediate stage between ICU 
                   and recovered) population time series
        """
        if None != Icpi:
            X  = data['X']
            Icp = X[:, Icpi[0]*self.M:Icpi[1]*self.M]
        else:
            X  = data['X']
            Icpi=self.readData['Icpi']
            Icp = X[:, Icpi[0]*self.M:Icpi[1]*self.M]
        return Icp


    def Im(self,  data, Imi=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             Ic : mortality time series
        """
        if None != Imi:
            X  = data['X']
            Im = X[:, Imi[0]*self.M:Imi[1]*self.M ]
        else:
            X  = data['X']
            Imi=self.readData['Imi']
            Im = X[:, Imi[0]*self.M:Imi[1]*self.M ]
        return Im


    def R(self,  data, Rind=None):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
             R: Removed population time series
             For SEAI8R: R=N(t)-(S+E+Ia+Is+Is'+Ih+Ih'+Ic+Ic')
        """
        if None != Rind:
            X = data['X']  
            R = self.population
        else:
            X = data['X']  
            Rind=self.readData['Rind']
            R = self.population
        for i in range(Rind):
            R  = R - X[:, i*self.M:(i+1)*self.M] 
        return R
   

    def Sx(self,  data, Sxi):
        """
        Parameters
        ----------
        data: Data dict

        Returns
        -------
            Generic compartment Sx 
        """
        X  = data['X']
        Im = X[:, Sxi[0]*self.M:Sxi[1]*self.M ]
        Sx = X[:, Sxi[0]*self.M:Sxi[1]*self.M ]
        return Sx



@cython.wraparound(False)
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.nonecheck(False)
cdef class SIkR(CommonMethods):
    """
    Susceptible, Infected, Removed (SIkR). Method of k-stages of I

    .. math::
        \dot{S_{i}}=-\lambda_{i}(t)S_{i},
    .. math::
        \dot{I}_{i}^{1}=k_{E}\gamma_{E}E_{i}^{k}-k_{I}\gamma_{I}I_{i}^{1},
    .. math::
        \dot{I}_{i}^{k}=k_{I}\gamma_{I}I_{i}^{(k-1)}-k_{I}\gamma_{I}I_{i}^{k}, 
    .. math::
        \dot{R}_{i}=k_{I}\gamma_{I}I_{i}^{k}.

    .. math::
        \lambda_{i}(t)=\\beta\sum_{j=1}^{M}\sum_{n=1}^{k}C_{ij}(t)\\frac{I_{j}^{n}}{N_{j}},
    ...

    Parameters
    ----------
    parameters: dict
        Contains the following keys:

        beta: float
            Rate of spread of infection.
        gI: float
            Rate of removal from infectives.
        kI: int
            number of stages of infection.
    M: int
        Number of compartments of individual for each class.
        I.e len(contactMatrix)
    Ni: np.array(M, )
        Initial number in each compartment and class
    """

    def __init__(self, parameters, M, Ni):
        self.beta  = parameters['beta']              # Infection rate
        self.gI    = parameters['gI']                # Removal rate of I
        self.kI    = parameters['kI']
        self.nClass= self.kI + 1

        self.N     = np.sum(Ni)
        self.M     = M
        self.Ni    = np.zeros( self.M, dtype=DTYPE)  
        self.Ni    = Ni

        self.CM    = np.zeros( (self.M, self.M), dtype=DTYPE)   # Contact matrix C
        self.dxdt  = np.zeros( (self.kI+1)*self.M, dtype=DTYPE) # Right hand side

        self.paramList = parameters
        self.readData = {'Ii':[1,self.kI+1], 'Rind':self.kI+1}
        self.population = self.Ni


    cpdef rhs(self, xt, tt):
        cdef:
            int N=self.N, M=self.M, i, j, jj, kI=self.kI
            double beta=self.beta, gI=self.kI*self.gI, rateS, lmda
            double [:] S    = xt[0  :M]
            double [:] I    = xt[M  :(kI+1)*M]
            double [:] Ni   = self.Ni
            double [:,:] CM = self.CM
            double [:] dxdt = self.dxdt

        for i in range(M):
            lmda=0
            for jj in range(kI):
                for j in range(M):
                    lmda += beta*(CM[i,j]*I[j+jj*M])/Ni[j]
            rateS = lmda*S[i]
            #
            dxdt[i]     = -rateS
            dxdt[i+M]   = rateS - gI*I[i]

            for j in range(kI-1):
                dxdt[i+(j+2)*M]   = gI*I[i+j*M] - gI*I[i+(j+1)*M]
        return


    def simulate(self, S0, I0, contactMatrix, Tf, Nf, Ti=0, integrator='odeint',
                 maxNumSteps=100000, **kwargs):
        """
        Simulates a compartment model given initial conditions,
        choice of integrator and other parameters. 
        Returns the time series data and parameters in a dict. 
        Internally calls the method 'simulator' of CommonMethods
        
        ...

        Parameters
        ----------
        S0: np.array
            Initial number of susceptables.
        I0: np.array
            Initial number of  infectives.
        contactMatrix: python function(t)
             The social contact matrix C_{ij} denotes the
             average number of contacts made per day by an
             individual in class i with an individual in class j
        Tf: float
            Final time of integrator
        Nf: Int
            Number of time points to evaluate.
        Ti: float, optional
            Start time of integrator. The default is 0.
        integrator: TYPE, optional
            Integrator to use either from scipy.integrate or odespy.
            The default is 'odeint'.
        maxNumSteps: int, optional
            maximum number of steps the integrator can take.
            The default is 100000.
        **kwargs: kwargs for integrator

        Returns
        -------
        dict
             X: output path from integrator,  t : time points evaluated at,
            'param': input param to integrator.

        """

        x0=np.concatenate((S0, I0))
        data = self.simulator(x0, contactMatrix, Tf, Nf, 
                              integrator, Ti, maxNumSteps, **kwargs)
        return data






cdef class Simulator:
    """
    Simulator for a deterministic time-since infection model

    Methods
    -------
    solve_Predictor_Corrector: 

    solve_Galerkin: 
    """


    cdef:
        readonly dict parameters
        readonly str method
        readonly str galerkinIntegrator 
        readonly list IC 
        readonly np.ndarray phi_alpha, p_alpha

    def __init__(self, parameters, method='Predictor_Corrector', galerkinIntegrator='odeint'):
        self.parameters         = parameters
        self.method             = method
        self.galerkinIntegrator = galerkinIntegrator 



    def get_IC(self):
        parameters = self.parameters
        M = parameters['M']
        Ni = parameters['Ni']
        Nc = parameters['Nc']
        Nk = parameters['Nk']
        contactMatrix = parameters['contactMatrix']
        Cij = contactMatrix(0)
    
    
        Np = sum(Ni)            #total population size
    
        T    = parameters['T']
        Td   = parameters['Td']
        Tc   = parameters['Tc']
        tsi  = parameters['tsi']
        beta = parameters['beta']

        if M == 1:
            S_0 = np.array([1])
            I_0 = .001*np.interp(np.linspace(0, 2 ,Nk),tsi, 2**(tsi/Td))
            Ic_0 = np.zeros((Nc,M))
            IC = [S_0, I_0, Ic_0]
            return IC

        A = np.matmul(np.diag(Ni),Cij);  
        A = np.matmul(A,np.diag(1/Ni)); 
        max_eig_A = np.max(np.real(np.linalg.eigvals(A)))
        sp   = np.linspace(0,T,1000); lam = np.log(2)/Td;  #Growth rate
        rs   = max_eig_A*np.trapz(np.exp(-lam*sp)*np.interp(sp,tsi,beta),sp)
        beta = beta/rs       #now beta has been rescaled to give the correct (dimensional) doubling time
        
        #Initial susceptible is the whole population
        S_0 = Ni
    
        #get I_0 from linear stability analysis
        #find the fastest growing linear mode
        A = np.matmul(np.diag(Ni),Cij)
        A = np.matmul(A,np.diag(1/Ni))
        sp = np.linspace(0,T,1000)
        lam = np.log(2)/Td;
        A = A*np.trapz(np.exp(-lam*sp)*np.interp(sp,tsi,beta),sp)
        w, v = np.linalg.eig(-np.identity(M) + A)

        ep = 0.001/T*Np/M

        #now identify the largest eigenvalue/eigenvector...
        pos = np.where(w == np.amax(w))
        pos = pos[0][0]
        lam = T/Td*np.log(2)
        s = np.linspace(-1,1,Nk)
        I_0 = ep*np.abs(np.real(np.outer(np.exp(-lam*(s+1)),v[:,pos])))
 
        #just set Ic_0 to zero -- these numbers are too small to matter
        Ic_0 = np.zeros((Nc,M))
    
        #rescale all population sizes
        Ni = Ni/Np
        I_0 = I_0/Np*Tc  #recall that this is a number density (number infected per unit tsi)
        S_0 = S_0/Np
        if Nc > 0:
            Ic_0 = Ic_0/Np*Tc
        IC = [S_0, I_0, Ic_0]
        return IC
    
    
    
    
    
    
    def solve_Predictor_Corrector(self, tstart=0, hybrid=False):
        ''' Predictor/Corrector is a finite difference method described in the TSI report, section 2.5
             It has good properties for speed and accuracy and should be preferred in most applications
             Notable disadvantage is a lack of flexibility in time-stepping -- you must increment by
             the same time step every time.  Function evaluations at intermediate times can be found by
             interpolation.
        '''
        M  = self.parameters['M']                  
        Ni = self.parameters['Ni']                  
        Nc = self.parameters['Nc']                   
        Nk = self.parameters['Nk']                   
        Tf = self.parameters['Tf']                   
        T  = self.parameters['T']                   
        Td = self.parameters['Tc']                   
        Td = self.parameters['Td']                   
       
        tsi       = self.parameters['tsi']
        beta      = self.parameters['beta']                  
        tsi_sc    = self.parameters['tsi_sc']

        p_alpha   = self.parameters['p_alpha']
        phi_alpha = self.parameters['phi_alpha']
        Np        = np.sum(Ni)
        Ni = Ni/Np                  
        
        contactMatrix = self.parameters['contactMatrix']
        def Cij_t(t):
            return np.matmul(contactMatrix(t*Tc),np.diag(1/Ni))


        Tc = T/2 #rescaling of time, and tsi 
        #first step is to rescale beta to a value consistent with the given Td
        Cij = contactMatrix(0)
        if M > 1:
            A = np.matmul(np.diag(Ni),Cij)
            A = np.matmul(A,np.diag(1/Ni))
        else:
            A = Cij
        max_eig_A = np.max(np.real(np.linalg.eigvals(A)))
        sp = np.linspace(0,T,1000)
        lam = np.log(2)/Td;  #Growth rate
        rs = max_eig_A*np.trapz(np.exp(-lam*sp)*np.interp(sp,tsi,beta),sp)
        beta = beta/rs       #now beta has been rescaled to give the correct (dimensional) doubling time

        #nondimensionalize beta:
        beta = beta*Tc
        tsi = tsi/Tc - 1
        
        #rescale phi_alpha based on re-scaling of tsi:
        tsi_sc = tsi_sc/Tc - 1
        
        #rescale end time
        Tf = Tf/Tc
        tstart = tstart/Tc

        #rescale phi_alpha
        if Nc > 0:
            phi_alpha = phi_alpha*Tc

        S_0, I_0, Ic_0 = self.IC 
   
 
        #set up the discretization in s
        s = np.linspace(-1,1,Nk)
        h = 2/(Nk - 1)
    
        #find the timesteps
        nt = int(np.round(Tf/h)) + 1
        t = h*np.linspace(0,nt-1,nt)

        #weighted betas for numerical integration
        beta_n = h*np.interp(s,tsi,beta)
        beta_n[[0, Nk-1]] = beta_n[[0, Nk-1]]/2
    
        #weighted phi_alpha for numerical integration
        phi_alpha_n = np.zeros((Nc, Nk))
        for i in range(Nc):
            phi_alpha_n[i,:] = h*np.interp(s,tsi_sc,phi_alpha[i,:])
            phi_alpha_n[i,[0,Nk-1]] = phi_alpha_n[i,[0,Nk-1]]/2
    
        #weights for generic trapezoid integration
        w = h*np.ones(Nk)
        w[[0, Nk-1]] = w[[0, Nk-1]]/2
    
        #initialize variables
        S = S_0 + 0
        I = I_0 + 0
        Ic = Ic_0 + 0
    
        #initialize output vectors:
        S_t = np.zeros((M, nt))
        I_t = np.zeros((M, nt))
        Ic_t = np.zeros((Nc,M,nt))
    
        #set their starting values:
        S_t[:,0] = S_0
        I_t[:,0] = np.matmul(w,I_0)
        Ic_t[:,:,0] = Ic_0
    
        #initialize a few variables
        dIc_dt_e = np.zeros((Nc,M))
        dIc_dt_i = np.zeros((Nc,M))

        for i in (1 + np.arange(nt-1)):
    
            #explicit time step
            dSdt_e = -np.matmul(np.matmul(np.diag(S),Cij_t(tstart + t[i-1])),np.matmul(beta_n,I))
            Sp = S + h*dSdt_e

            for j in range(Nc):
                dIc_dt_e[j,:] = np.matmul(phi_alpha_n[j,:],I)*p_alpha[j,:]

            I[1:Nk,:] = I[0:(Nk-1),:]
            I[0,:] = -dSdt_e

            #'implicit' step
            dSdt_i = -np.matmul(np.matmul(np.diag(Sp),Cij_t(tstart + t[i])),np.matmul(beta_n,I))
            S = S + h/2*(dSdt_e + dSdt_i)

            for j in range(Nc):
                dIc_dt_i[j,:] = np.matmul(phi_alpha_n[j,:],I)*p_alpha[j,:]
    
            Ic = Ic + h/2*(dIc_dt_e + dIc_dt_i)
    
            I[0,:] = -dSdt_i
    
            #remember this timestep
            S_t[:,i]     = S
            I_t[:,i]     = np.matmul(w,I)
            Ic_t[:,:, i] = Ic 
    
        if not hybrid:
            return t, S_t, I_t, Ic_t
        else:
            return t, S_t, I_t, Ic_t, [S, I, Ic]



    
    def solve_Galerkin(self, atol=1e-4, rtol=1e-3, tstart=0, hybrid=False):
        '''The Galerkin method is defined in the TSI report, section 2.6
         It spectral accuracy in s and allows for adatptive timestepping in t
         For constant contact matrix, use 'odeint', otherise use 'Crank Nicolson'
             -Notable advantage over predictor/corrector is flexibility in time-stepping
             -Notable disadvantages include:
              (1) ill-suited to non-smooth dynamics (like most spectral methods)
              (2) must be solved as DAE when contact matrix is time-dependent (slow)
        
        For most practical purposes, we regard predictor/corrector as the preferred choice.
        '''
        M  = self.parameters['M']                  
        Ni = self.parameters['Ni']                   
        Nc = self.parameters['Nc']                   
        Nk = self.parameters['Nk']                   
        NL = self.parameters['NL']                   
        Tf = self.parameters['Tf']                   
        Tc = self.parameters['Tc']
        Td = self.parameters['Td']
        T  = self.parameters['T']                   
       
        tsi       = self.parameters['tsi']
        beta      = self.parameters['beta']                  
        tsi_sc    = self.parameters['tsi_sc']                  
        
        p_alpha   = self.parameters['p_alpha']
        phi_alpha = self.parameters['phi_alpha']

        galerkinIntegrator = self.galerkinIntegrator

        Np        = np.sum(Ni)
        Ni = Ni/Np                  
        
        contactMatrix = self.parameters['contactMatrix']
        def Cij_t(t):
            return np.matmul(contactMatrix(t*Tc),np.diag(1/Ni))


        Tc = T/2 #rescaling of time, and tsi 
        #first step is to rescale beta to a value consistent with the given Td
        Cij = contactMatrix(0)
        if M > 1:
            A = np.matmul(np.diag(Ni),Cij)
            A = np.matmul(A,np.diag(1/Ni))
        else:
            A = Cij
        max_eig_A = np.max(np.real(np.linalg.eigvals(A)))
        sp = np.linspace(0,T,1000)
        lam = np.log(2)/Td;  #Growth rate
        rs = max_eig_A*np.trapz(np.exp(-lam*sp)*np.interp(sp,tsi,beta),sp)
        beta = beta/rs       #now beta has been rescaled to give the correct (dimensional) doubling time
   
        #nondimensionalize beta:
        beta = beta*Tc
        tsi = tsi/Tc - 1
        
        #rescale phi_alpha based on re-scaling of tsi:
        tsi_sc = tsi_sc/Tc - 1
        
        #rescale end time
        Tf = Tf/Tc
        tstart = tstart/Tc

        #rescale phi_alpha
        if Nc > 0:
            phi_alpha = phi_alpha*Tc


        galerkinIntegrator  = self.galerkinIntegrator


        S_0, I_0, Ic_0 = self.IC
    
        #set up the discretization in s
        s = np.linspace(-1,1,1000)
        sk = np.linspace(-1,1,Nk)
    
    
        #initialize first timestep
        h = 1/Nk
    
        #weighted betas for numerical integration
        beta_n = np.zeros(NL)
        for i in range(NL):
            Pn = legendre(i)
            beta_n[i] = np.trapz(Pn(s)*np.interp(s,tsi,beta),s)
    
        #weighted phi_alpha for numerical integration
        phi_alpha_n = np.zeros((Nc, NL))
        for i in range(Nc):
            for j in range(NL):
                Pn = legendre(j)
                phi_alpha_n[i,j] = np.trapz(Pn(s)*np.interp(s,tsi_sc,phi_alpha[i,:]),s)
    
        #advection matrix
        b_nk = np.zeros((NL,NL))
        for i in range(NL):
            for j in range(NL):
                if np.mod(i + j,2) == 1 and j > i:
                    b_nk[i,j] = 2*i + 1
    
        #initialize variables
        S = S_0 + 0
        I = np.zeros((NL,M))
        for i in range(NL):
            Pn = legendre(i)
            for j in range(M):
                I[i,j] = np.trapz(Pn(s)*np.interp(s,sk,I_0[:,j]),s)*(2*i + 1)/2
    
        Ic = Ic_0 + 0
    
        #set their starting values:
        S_t = S_0
        I_t = 2*I[0]
        Ic_t = Ic_0
    
        if galerkinIntegrator == 'Crank Nicolson':
    
            #define a function to spit out time derivatives
            def get_dxdt(x, t):
                Cij = Cij_t(tstart + t)
                S = x[:M]
                I = np.transpose(np.reshape(x[M:],(M,NL)))
                dxdt = np.zeros(len(x))
                for i in range(M):
                    dxdt[i] = -S[i]*np.matmul(Cij[i,:],np.matmul(beta_n,I))
                for i in range(M):
                    dxdt[M + i*NL + np.array(range(NL))] = -np.matmul(b_nk,I[:,i])
                return dxdt
    
            #define a function to give 'residuals' based on current estimate of next time step
            def get_res(x, xp, t, dt):
                #Cij = Cij_t(tstart + t)
                S = x[:M]
                I = np.transpose(np.reshape(x[M:],(M,NL)))
                Sp = xp[:M]
                Ip = np.transpose(np.reshape(xp[M:],(M,NL)))
                dxdt_e = get_dxdt(x,t)
                dxdt_i = get_dxdt(xp,t+dt)
                res = xp - x - dt/2*(dxdt_e + dxdt_i)
                for i in range(M):
                    res[M + i*NL + NL - 1] = np.matmul((-1.)**np.array(range(NL)),Ip[:,i]) + dxdt_i[i]
                return res
    
            #define a function to compute Jacobians.
            def get_J(x, t, dt):
                Cij = Cij_t(tstart + t + dt)
                S = x[:M]
                I = np.transpose(np.reshape(x[M:],(M,NL)))
                J = np.zeros((len(x),len(x)))
                for i in range(M):
                    J[i,i] = 1 + dt/2*np.matmul(Cij[i,:],np.matmul(beta_n,I))
                for i in range(M):
                    for j in range(M):
                        for k in range(NL):
                            J[i,M + j*NL + k] = dt/2*S[i]*Cij[i,j]*beta_n[k]
                for i in range(M):
                    q = np.identity(NL) + dt/2*b_nk
                    for j in range(NL - 1):
                        J[M + NL*i + j, (i*NL + M):(i*NL + M + NL)] = q[j,:]
                for i in range(M):
                    J[M + i*NL + NL - 1,i] = -np.matmul(Cij[i,:],np.matmul(beta_n,I))
                    J[M + i*NL + NL - 1,(M + i*NL):(M + i*NL + NL)] = (-1)**np.arange(NL)
                    for j in range(M):
                        J[M + i*NL + NL - 1,(M + j*NL):(M + j*NL + NL)] += - S[i]*Cij[i,j]*beta_n
                return J
    
            t_t = np.array([0])
            t = 0
    
            x = np.append(S,np.transpose(I))
            xh2 = x + 0;
            res = get_res(x, xh2, t, h)
    
            def get_next_step(x,x0,t,dt):
                xp = x0 + 0
                res = get_res(x,xp,t,dt)
                err = np.amax(np.abs(res))
                count = 0
                maxiter = 100
                while np.greater(err,0.000001) and count < maxiter:
                    J = get_J(xp,t,dt)
                    dx = np.linalg.solve(J,res)
                    xp += -dx
                    res = get_res(x, xp, t, dt)
                    err = np.amax(np.abs(res))
                    count += 1
                if count == maxiter:
                    print('solver maxed out')
                return xp
    
            while t < Tf:
                e_abs = 100
                e_rel = 100
                count = 0
                maxiter = 20
                x = xh2
    
                while (e_abs > atol or e_rel > rtol) and count < maxiter:
                    count+=1
                    if t + h > Tf:
                        h = Tf - t
                    #full step
                    xf  = get_next_step(x,x,t,h)
    
                    #half steps
                    xh1 = get_next_step(x,1/2*(x + xf),t, h/2)
                    xh2 = get_next_step(xh1, xf,t +h/2, h/2)
    
                    #compare predictions
                    err = np.abs(xh2 - xf)
                    e_abs = np.amax(err)
                    e_rel = np.amax(err/np.abs(xh2))
    
                    #adaptive time stepping
                    #reduce time step if not meeting error tolerances
                    if e_abs > atol:
                        h = 0.8*h*(atol/e_abs)**.5
                    elif e_rel > rtol:
                        h = 0.8*h*(rtol/e_rel)**.5
    
                #increment
                t_t = np.append(t_t,np.array([t + h/2, t + h]))
                t +=h
                count +=1
    
                #remember some of the results
                S_t = np.append(S_t, xh1[:M])
                S_t = np.append(S_t, xh2[:M])
                I_t = np.append(I_t, 2*xh1[M + NL*np.arange(M)])
                I_t = np.append(I_t, 2*xh2[M + NL*np.arange(M)])
                #subclasses of infecteds:
                Ii = np.transpose(np.reshape(xh1[M:],(M,NL)))
                Ie = np.transpose(np.reshape(x[M:],(M,NL)))
                dIc_dt_i = np.zeros((Nc,M))
                dIc_dt_e = np.zeros((Nc,M))
                for j in range(Nc):
                    dIc_dt_i[j,:] = np.matmul(phi_alpha_n[j,:],Ii)*p_alpha[j,:]
                    dIc_dt_e[j,:] = np.matmul(phi_alpha_n[j,:],Ie)*p_alpha[j,:]
                Ic += h/4*(dIc_dt_e + dIc_dt_i)
                if t == h:
                    Ic_t = np.concatenate(([Ic_0], [Ic]))
                else:
                    Ic_t = np.concatenate((Ic_t, [Ic]))
    
                Ie = Ii + 0
                Ii = np.transpose(np.reshape(xh2[M:],(M,NL)))
                for j in range(Nc):
                    dIc_dt_i[j,:] = np.matmul(phi_alpha_n[j,:],Ii)*p_alpha[j,:]
                    dIc_dt_e[j,:] = np.matmul(phi_alpha_n[j,:],Ie)*p_alpha[j,:]
                Ic += h/4*(dIc_dt_e + dIc_dt_i)
                Ic_t = np.concatenate((Ic_t, [Ic]))
    
                #adaptive time stepping
                #bump time step if well below threshold.
                if e_abs == 0:
                    h = 2*h
                else:
                    if e_rel/rtol >= e_abs/atol:
                        h = 0.8*h*(rtol/e_rel)**.5
                    else:
                        h = 0.8*h*(atol/e_abs)**.5
    
                if count == maxiter:
                    print('CN solver failed')
                    t = Tf
    
            #reshape the output of infecteds class:
            Ic_t_reshape = np.zeros((Nc,M,len(t_t)))
            for i in range(len(t_t)):
                Ic_t_reshape[:,:,i] = Ic_t[i,:,:]
    
            if not hybrid:
                return t_t, np.transpose(np.reshape(S_t,(len(t_t),M))), np.transpose(np.reshape(I_t,(len(t_t),M))), Ic_t_reshape
            else:
                #compute IC for next run from final point of current run.
                S_0 = xh2[:M]
                Ic_0 = Ic
                I_0 = np.zeros((Nk,M))
                I_end = np.reshape(xh2[M:],(M,NL))
                for i in range(M):
                    for j in range(NL):
                        Pn = legendre(j)
                        I_0[:,i] += Pn(sk)*I_end[i,j]
                return t_t, np.transpose(np.reshape(S_t,(len(t_t),M))), np.transpose(np.reshape(I_t,(len(t_t),M))), Ic_t_reshape, [S_0, I_0, Ic_0]
    
        elif galerkinIntegrator == 'odeint':
            #compute time derivative w. explicit treatment of BC
            def get_dxdt(x, t):
                Cij = Cij_t(tstart + t)
    
                S = x[:M]
                I = np.zeros((NL,M))
                I[:(NL-1), :] = np.transpose(np.reshape(x[M:(M + M*(NL-1))],(M,NL-1)))
    
                #solve for the highest Legendre polynomial coefficients:
                A = np.identity(M)*(-1)**(NL-1) - np.matmul(np.diag(S),Cij)*beta_n[NL-1]
                b = np.zeros(M)
                for i in range(M):
                    b[i] = S[i]*np.matmul(Cij[i,:],np.matmul(beta_n,I)) - np.sum(I[:,i]*(-1)**np.arange(NL))
                IN = np.linalg.solve(A,b)
                I[NL-1,:] = IN
    
                #compute derivs for S, I:
                dxdt = np.zeros(len(x))
                for i in range(M):
                    dxdt[i] = -S[i]*np.matmul(Cij[i,:],np.matmul(beta_n,I))
                for i in range(M):
                    dxdt[M + i*(NL-1) + np.array(range(NL-1))] = -np.matmul(b_nk[:(NL-1),:],I[:,i])
    
                #compute derivs for subclasses of infecteds:
                dIc_dt = np.zeros((Nc,M))
                for i in range(Nc):
                    dIc_dt[i,:] = np.matmul(phi_alpha_n[i,:],I)*p_alpha[i,:]
                dxdt[(M + (M)*(NL-1)):] = np.reshape(dIc_dt,(1,Nc*M))
    
                return dxdt
    
            #set initial condition
            x0 = np.append(S,np.transpose(I[:(NL-1),:]))
            x0 = np.append(x0,Ic)
    
            #choose times for reporting output
            t = np.linspace(0, Tf, int(Tf*2*Nk));
            nt = len(t)
    
            #get solution
            u = odeint(get_dxdt, x0, t,rtol = rtol, atol = atol)
    
            #transform solution to get outputs
            S_t = np.transpose(u[:,:M])
            I_t = 2*np.transpose(u[:,M + (NL - 1)*np.arange(M)])
            Ic_t = np.zeros((Nc,M,nt))
            for i in range(nt):
                Ic_t[:,:,i] = np.reshape(u[i,(M + (NL-1)*(M)):],(Nc,M))
    
            if not hybrid:
                return t, S_t, I_t, Ic_t
            else:
                #compute IC for next run from final point of current run.
                S_0 = S_t[:,nt-1]
                Ic_0 = Ic_t[:,:,nt-1]
                I_0 = np.zeros((Nk,M))
                I_end = np.zeros((M, NL))
                I_end[:,:(NL-1)] = np.reshape(u[nt-1,M:(M + (NL-1)*M)],(M,NL-1))
    
                A = np.identity(M)*(-1)**(NL-1) - np.matmul(np.diag(S_t[:,nt-1]),Cij_t(t[nt - 1]))*beta_n[NL-1]
                b = np.zeros(M)
                for i in range(M):
                    b[i] = S_t[i,nt-1]*np.matmul(Cij_t(t[nt - 1])[i,:],np.matmul(beta_n,np.transpose(I_end))) - np.matmul(I_end[i,:],(-1)**np.arange(NL))
                IN = np.linalg.solve(A,b)
                I_end[:,NL-1] = IN
    
                for i in range(M):
                    for j in range(NL):
                        Pn = legendre(j)
                        I_0[:,i] += Pn(sk)*I_end[i,j]
                return t, S_t, I_t, Ic_t, [S_0, I_0, Ic_0]
        else:
            print('please choose a valid method for solving Galerkin -- Crank Nicolson or odeint')


    def solve_Predictor_CorrectorQ(self):
        ''' This is a predictor corrector method specifically for epidemics with two quarantine classes:
            (1) Voluntary quarantines -- people move in and out based on the current advice/symptoms, no concern for history
            (2) Assigned quarantines -- whether a person is quarantined or not depends on testing capabilities in the past
            Voluntary quarantines are all assumed to end before time T, whereas assigned quarantines are assumed to all last
            beyond time T.
        '''
        M  = self.parameters['M']                  
        Ni = self.parameters['Ni']                  
        Nc = self.parameters['Nc']                   
        Nk = self.parameters['Nk']                   
        Tf = self.parameters['Tf']                   
        T  = self.parameters['T']                   
        Td = self.parameters['Tc']                   
        Td = self.parameters['Td']                   
       
        tsi       = self.parameters['tsi']
        beta      = self.parameters['beta']                  
        tsi_sc    = self.parameters['tsi_sc']

        p_alpha   = self.parameters['p_alpha']
        phi_alpha = self.parameters['phi_alpha']
        Np        = np.sum(Ni)
        Ni = Ni/Np

        #unbundle quarantine parameters:
        tsiQ = self.parameters['tsiQ']
        TQ = self.parameters['TQ']
        PhiQV = self.parameters['PhiQV']
        phiQA = self.parameters['phiQA']
        Phi_QA_0 = self.parameters['Phi_QA_0']
        pQV = self.parameters['pQV_t']
        pQA = self.parameters['pQA_t']

        contactMatrixQ = self.parameters['CijQ_t']
        def CijQ_t(t):
            return np.matmul(contactMatrixQ(t*Tc),np.diag(1/Ni))
        def pQV_t(t):
            return pQV(t*Tc)
        def pQA_t(t):
            return pQA(t*Tc)
        
        contactMatrix = self.parameters['contactMatrix']
        def Cij_t(t):
            return np.matmul(contactMatrix(t*Tc),np.diag(1/Ni))


        Tc = T/2 #rescaling of time, and tsi 
        #first step is to rescale beta to a value consistent with the given Td
        Cij = contactMatrix(0)
        if M > 1:
            A = np.matmul(np.diag(Ni),Cij)
            A = np.matmul(A,np.diag(1/Ni))
        else:
            A = Cij
        max_eig_A = np.max(np.real(np.linalg.eigvals(A)))
        sp = np.linspace(0,T,1000)
        lam = np.log(2)/Td;  #Growth rate
        rs = max_eig_A*np.trapz(np.exp(-lam*sp)*np.interp(sp,tsi,beta),sp)
        beta = beta/rs       #now beta has been rescaled to give the correct (dimensional) doubling time
   
        #nondimensionalize beta:
        beta = beta*Tc
        tsi = tsi/Tc - 1
        
        #rescale phi_alpha based on re-scaling of tsi:
        tsi_sc = tsi_sc/Tc - 1
        tsiQ = tsiQ/Tc - 1
        
        #rescale end time
        Tf = Tf/Tc
        TQ = TQ/Tc

        #rescale phi_alpha
        if Nc > 0:
            phi_alpha = phi_alpha*Tc
        phiQA = phiQA*Tc

        S_0, I_0, Ic_0 = self.IC

        #scalings are done.  Begin setup.
        #set up the discretization in s
        s = np.linspace(-1,1,Nk)
        h = 2/(Nk - 1)
    
        #find the timesteps
        nt = int(np.round(Tf/h)) + 1
        t = h*np.linspace(0,nt-1,nt)
        
        #weighted betas for numerical integration
        beta_n = h*np.interp(s,tsi,beta)
        beta_n[[0, Nk-1]] = beta_n[[0, Nk-1]]/2
    
        #weighted phi_alpha for numerical integration
        phi_alpha_n = np.zeros((Nc, Nk))
        for i in range(Nc):      
            phi_alpha_n[i,:] = h*np.interp(s,tsi_sc,phi_alpha[i,:])
            phi_alpha_n[i,[0,Nk-1]] = phi_alpha_n[i,[0,Nk-1]]/2
    
        #weights for generic trapezoid integration
        w = h*np.ones(Nk)
        w[[0, Nk-1]] = w[[0, Nk-1]]/2
        
        #initialize variables
        S = S_0 + 0
        I = I_0 + 0
        Ic = Ic_0 + 0
            
        #initialize output vectors:
        S_t = np.zeros((M, nt))
        I_t = np.zeros((M, nt))
        Ic_t = np.zeros((Nc,M,nt))
        
        #set their starting values:
        S_t[:,0] = S_0    
        I_t[:,0] = np.matmul(w,I_0)
        Ic_t[:,:,0] = Ic_0

        #initialize a few variables
        dIc_dt_e = np.zeros((Nc,M))
        dIc_dt_i = np.zeros((Nc,M))
    
        #Voluntary quarantines
        Phi_QV_n = h*np.interp(s,tsiQ,PhiQV)

        #Assigned quarantines
        phi_QA_n = h*np.interp(s,tsiQ,phiQA)   #further processing needed
        Phi_QA_n = Phi_QA_0
        
        #initialize variables/lists
        IQV = np.matmul(w,np.outer(Phi_QV_n,pQV_t(0))*I_0)
        IQA = 0
        IQV_t = np.zeros((M,nt))
        IQA_t = np.zeros((M,nt))
        IQV_t[:,0] = IQV
        IQA_t[:,0] = IQA
           
    
        for i in (1 + np.arange(nt-1)):
            ######################
            # explicit time step #
            ######################
            
            IQA = Phi_QA_n*I
            IQV = np.outer(Phi_QV_n,pQV(t[i-1]))*I
            IQ = IQA + IQV
            dSdt_e = -np.matmul(np.matmul(np.diag(S), Cij_t(t[i-1])),np.matmul(beta_n,I-IQ)) + \
                     -np.matmul(np.matmul(np.diag(S),CijQ_t(t[i-1])),np.matmul(beta_n,  IQ))
               
            Sp = S + h*dSdt_e
            
            for j in range(Nc):
                dIc_dt_e[j,:] = np.matmul(phi_alpha_n[j,:],I)*p_alpha[j,:]
            
                    
            #update assigned quarantine profile:
            dPhi_QA_dt = 1/2*(np.outer(phi_QA_n[1:],      pQA_t(t[i-1])) +\
                              np.outer(phi_QA_n[:(Nk-1)], pQA_t(t[i-0])))
            dIQA_dt = 1/2*(np.outer(phi_QA_n[1:],      pQA_t(t[i-1]))*I[1:     ,:] +\
                           np.outer(phi_QA_n[:(Nk-1)], pQA_t(t[i-0]))*I[:(Nk-1),:])
            Phi_QA_n[1:,:] = Phi_QA_n[:(Nk-1),:] + h*dPhi_QA_dt
            
            #update infected profile
            I[1:Nk,:] = I[0:(Nk-1),:]
            I[0,:] = -dSdt_e
                                                            
            ##################
            #'implicit' step #
            ##################                                                    
            IQA = Phi_QA_n*I
            IQV = np.outer(Phi_QV_n,pQV(t[i]))*I
            IQ = IQA + IQV
            dSdt_i = -np.matmul(np.matmul(np.diag(Sp), Cij_t(t[i])),np.matmul(beta_n,I-IQ)) + \
                     -np.matmul(np.matmul(np.diag(Sp),CijQ_t(t[i])),np.matmul(beta_n,  IQ))
    
            S = S + h/2*(dSdt_e + dSdt_i)
            
            for j in range(Nc):
                dIc_dt_i[j,:] = np.matmul(phi_alpha_n[j,:],I)*p_alpha[j,:]
            
            Ic = Ic + h/2*(dIc_dt_e + dIc_dt_i)
    
            I[0,:] = -dSdt_i
            
            #remember this timestep    
            S_t[:,i]     = S    
            I_t[:,i]     = np.matmul(w,I)
            Ic_t[:,:, i] = Ic
            IQV_t[:,i]   = np.matmul(w,IQV)
            IQA_t[:,i]   = IQA_t[:,i-1] + h*np.sum(dIQA_dt,0)
        

        #back calculate the population currently under quarantine:
        tb = t[nt-1]; count = nt-1
        while (tb - h) > TQ:
            for i in range(M):
                IQA_t[i,count] += -np.interp(tb - TQ, t ,IQA_t[i,:])
            count += - 1
            tb += -h
            
        return t, S_t, I_t, Ic_t, IQV_t, IQA_t                      




    def simulate(self, IC, atol=1e-4, rtol=1e-3):
        self.IC = IC
        M  = self.parameters['M']                  
        T  = self.parameters['T']                   
        Nc = self.parameters['Nc']                   
        Nk = self.parameters['Nk']                   
        Tf = self.parameters['Tf']
        Tc = T/2                   
        
        tsi       = self.parameters['tsi']
        beta      = self.parameters['beta']                  
        beta      = self.parameters['beta']                  
        tsi_sc    = self.parameters['tsi_sc']                  
        
        method    = self.method
        galerkinIntegrator= self.galerkinIntegrator
        phi_alpha = self.parameters['phi_alpha']
        p_alpha   = self.parameters['p_alpha']                 
        
        contactMatrix = self.parameters['contactMatrix']

        if 'TQ' in self.parameters:
            t, S_t, I_t, Ic_t, IQV_t, IQA_t = self.solve_Predictor_CorrectorQ()
            data = {'t':t*T/2, 'S_t':S_t, 'I_t':I_t, 'Ic_t':Ic_t, 'IQV_t':IQV_t, 'IQA_t':IQA_t}
            return data
        
        if method == 'Predictor_Corrector':
            t, S_t, I_t, Ic_t = self.solve_Predictor_Corrector()

        elif method == 'Galerkin':
            t, S_t, I_t, Ic_t = self.solve_Galerkin(atol, rtol)
        
        elif method == 'Hybrid':
            tc = 0
            tswap = self.parameters['tswap']
            count = 0
            while tc < Tf:
                
                #run the next simulation
                if tc < tswap[count]:
                    tstep = tswap[count] - tc
                    self.parameters['Tf'] = tstep
                    sol = self.solve_Galerkin(atol, rtol, tc, True)
                else:
                    tstep = T
                    self.parameters['Tf'] = tstep
                    sol = self.solve_Predictor_Corrector(tc, True)
                    if count < len(tswap)-1:
                        count += 1
                
                #unpack results
                if count == 0 or (count == 1 and tc == 0):
                    t, S_t, I_t, Ic_t, self.IC = sol
                else:
                    t  = np.concatenate((t,tc/Tc + sol[0]))
                    nt = len(t); S_t_new = np.zeros((M,nt)); I_t_new = np.zeros((M,nt)); Ic_t_new = np.zeros((Nc,M,nt))
                    for i in range(M):
                        S_t_new[i,:] = np.append(S_t[i,:],sol[1][i,:])
                    for i in range(M):
                        I_t_new[i,:] = np.append(I_t[i,:],sol[2][i,:])
                    for i in range(Nc):
                        for j in range(M):
                            Ic_t_new[i,j,:] = np.append(Ic_t[i,j,:],sol[3][i,j,:])
                    self.IC = sol[4]
                    S_t, I_t, Ic_t = [S_t_new, I_t_new, Ic_t_new]
                
                #prepare for next loop
                tc = tc + tstep
                #print(IC_t)

        data = {'t':t*T/2, 'S_t':S_t, 'I_t':I_t, 'Ic_t':Ic_t}
        return data    
